/*
 * ParticlesTransformFeedback.cpp - Version 1.0 - May 22, 2020
 *
 * Example program illustrating of transform feedback in
 *   in a shader program to animate a system of particles
 *
 * Author: Sam Buss
 *
 * Software accompanying POSSIBLE SECOND EDITION TO the book
 *		3D Computer Graphics: A Mathematical Introduction with OpenGL,
 *		by S. Buss, Cambridge University Press, 2003.
 *
 * Software is "as-is" and carries no warranty.  It may be used without
 *   restriction, but if you modify it, please change the filenames to
 *   prevent confusion between different versions.
 * Bug reports: Sam Buss, sbuss@ucsd.edu.
 * Web page: http://math.ucsd.edu/~sbuss/MathCG2
 */

// Use space to toggle what image is shown.
// Use Escape or 'X' or 'x' to exit.

// These libraries are needed to link the program (Visual Studio specific)
#pragma comment(lib,"opengl32.lib")
#pragma comment(lib,"glu32.lib")
#pragma comment(lib,"glfw3.lib")
#pragma comment(lib,"glew32s.lib")
#pragma comment(lib,"glew32.lib")

// Use the static library (so glew32.dll is not needed):
#define GLEW_STATIC
#include <GL/glew.h> 
#include <GLFW/glfw3.h>

#include "GlShaderMgr.h"
bool check_for_opengl_errors();     // Function prototype (should really go in a header file)

// Enable standard input and output via printf(), etc.
// Put this include *after* the includes for glew and GLFW!
#include <stdio.h>

// ********************
// Animation controls and state infornation
// ********************

// There will be an NxN array of particles.
const int N = 2000;

// These two variables control whether running or paused.
bool spinMode = false;
bool singleStep = false;
bool firstTimeRendering = true;
int swapIndex = 0;              // Swapping between transform feedback buffers

unsigned int windowWidth;   // Size of viewport in pixels
unsigned int windowHeight;

// ************************
// General data helping with setting up VAO (Vertex Array Objects)
//    and Vertex Buffer Objects.
// ***********************

// We make two each of the VAO, VBO and TF objects. 
// This is so we can alternate between them when rendering
unsigned int myVAO[2];  // Vertex Array Object - holds info about an array of vertex data;
unsigned int myVBO[2];  // Vertex Buffer Object - holds an array of data
unsigned int myTF[2];  // Tranform Feedback - container for transform feedback

// We create one shader program: it consists of a vertex shader and a fragment shader
unsigned int shaderParticles;
const unsigned int vertPos_loc = 0;    // Corresponds to "location = 0" in the verter shader definition
const unsigned int vertVel_loc = 1;    // Corresponds to "location = 1" in the verter shader definition
const unsigned int vertID_loc = 2; // Corresponds to "location = 2" in the verter shader definition

// Locations of uniform variables in the current shader program
unsigned int mousePos_loc;
unsigned int RunSim_loc;
unsigned int ParticlePointSize_loc;

// *************************
// mySetupGeometries defines the scene data, especially vertex  positions and colors.
//    - It also loads all the data into the VAO's and VBO's.
// This routine is only called once to initialize the data.

void ResetParticles(); // Function prototype

// *************************
void mySetupGeometries() {

	// We do not use the Projection or ModelView matrices. 
    // Hence, all x, y, z positions should be in the range [-1,1].

    // Allocate Vertex Array Objects (VAOs) and Vertex Buffer Objects (VBOs) and two Transform Feedback's.
    glGenVertexArrays(2,&myVAO[0]);
    glGenBuffers(2, &myVBO[0]);
    glGenTransformFeedbacks(2, &myTF[0]);

    // Set the vertex attribute information in both VAO's.
    // myVAO[0] uses myVBO[0].  myVAO[1] uses myVBO[1].
    int numWords = 5 * (1 + N * N);
    for (int i = 0; i < 2; i++) {
        glBindVertexArray(myVAO[i]);
        glBindBuffer(GL_ARRAY_BUFFER, myVBO[i]);
        // Allocate buffer size for myVBO[i].
        glBufferData(GL_ARRAY_BUFFER, numWords * sizeof(float), (void*)0, GL_DYNAMIC_DRAW);
        glVertexAttribPointer(vertPos_loc, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)0);
        glEnableVertexAttribArray(vertPos_loc);
        glVertexAttribPointer(vertVel_loc, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(2 * sizeof(float)));
        glEnableVertexAttribArray(vertVel_loc);
        glVertexAttribIPointer(vertID_loc, 1, GL_INT, 5 * sizeof(float), (void*)(4 * sizeof(float)));
        glEnableVertexAttribArray(vertID_loc);
    }
    ResetParticles();

    // Let the two transform feedback buffers point to the two VBO buffers.
    glBindTransformFeedback(GL_TRANSFORM_FEEDBACK, myTF[0]);
    glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, 0, myVBO[0]);
    glBindTransformFeedback(GL_TRANSFORM_FEEDBACK, myTF[1]);
    glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, 0, myVBO[1]);

    // Just for safety, unbind all the buffers.
    glBindTransformFeedback(GL_TRANSFORM_FEEDBACK, 0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

void ResetParticles() {
    // Load the VBO[0] data, setup vertex attribute information in both VAO's
    // The first entry in the VBO is the target position (tracks the mouse)
    // The remaining NxN entries are the particles.

    // Each particle has: position - vec2 (two floats)
    //                    velocity - vec2 (two floats)
    //                    particle ID - 32 bit integer,  >=0 for particles, -1 for the target.
    // initParticles holds all these in one array of floats, 
    //         but every fifth entry is actually an integer.
    int numWords = 5 * (1 + N * N);
    float* initParticles = new float[numWords];
    initParticles[0] = initParticles[1] = initParticles[2] = initParticles[3] = 0.0;
    *(GLint*)(initParticles + 4) = -1;
    float delta = 2.0f / (float)N;  // Spacing between particles
    for (int i = 0; i < N * N; i++) {
        float* thisParticle = initParticles + 5 * (i + 1);
        thisParticle[0] = -1.0f + (0.5f + (float)(i % N)) * delta;  // Initial x position
        thisParticle[1] = -1.0f + (0.5f + (float)(i / N)) * delta;  // Initial y position
        thisParticle[2] = thisParticle[3] = 0.0f;    // Initial velocity is zero
        *(GLint*)(thisParticle + 4) = i + 1;              // Particle ID
    }

    glBindBuffer(GL_ARRAY_BUFFER, myVBO[0]);
    // Load data into myVBO[0]. Just allocate buffer size for myVBO[1].
    glBufferData(GL_ARRAY_BUFFER, numWords * sizeof(float), initParticles, GL_DYNAMIC_DRAW);

}

// *************************************
// Main routine for rendering the scene
// myRenderScene() is called every time the scene needs to be redrawn.
// mySetupGeometries() has already created the vertex and buffer objects, etc.
// setup_shaders() has already created the shader programs.
// *************************************
void myRenderScene() {

	// Clear the rendering window
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// This very simple shader program is used for all the items.
	glUseProgram(shaderParticles);

    // Render using the VAO and VBO for swapIndex
    glBindVertexArray(myVAO[swapIndex]);
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!

    // Transform feedback goes into the other VBO.
    glBindTransformFeedback(GL_TRANSFORM_FEEDBACK, myTF[1-swapIndex]);
    glBeginTransformFeedback(GL_POINTS);
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!

     if ( firstTimeRendering ) {
        glDrawArrays(GL_POINTS, 0, 1+N*N);   // Must use glDrawArrays the first time to specify number of points
        firstTimeRendering = !spinMode;
    }
    else {
        // This uses the size of the Transform Feedback data to specify the number of points <numPoints>.
        // Apart from that, the next line is equivalent to using glDrawArrays(GL_POINTS, 0, <numPoints>).
        glDrawTransformFeedback(GL_POINTS, myTF[swapIndex]);
    }
    glEndTransformFeedback();
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!

    //float* feedbackFloat = new float[5*(1+N*N)];
    //glGetBufferSubData(GL_TRANSFORM_FEEDBACK_BUFFER, 0, 4*5*(1+N*N), feedbackFloat);

    if (spinMode) {
        swapIndex = 1 - swapIndex;      // Alternate between buffers
    }
    if (spinMode && singleStep) {
        spinMode = false;       // If in single step mode, turn off future animation
        glUniform1i(RunSim_loc, 0);
    }
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
   
    // Unbind (optional)
    glBindVertexArray(0);
    glBindTransformFeedback(GL_TRANSFORM_FEEDBACK, 0);

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

void my_setup_SceneData() {
    mySetupGeometries();

    GlShaderMgr::LoadShaderSource("ParticlesTransformFeedback.glsl");

    unsigned int vertexShader = GlShaderMgr::CompileShader("vs1_PTF");
    unsigned int fragmentShader = GlShaderMgr::CompileShader("fragmentShader_ColorOnly");
    unsigned int shaders[2] = { vertexShader, fragmentShader };
    shaderParticles = glCreateProgram();
    glAttachShader(shaderParticles, vertexShader);
    glAttachShader(shaderParticles, fragmentShader);
    // The order of variables in this array is the order they will be put in the transform buffer object (TBO)
    const char* feedbackValues[5] = { "posX", "posY", "velX", "velY", "vertID" };
    glTransformFeedbackVaryings(shaderParticles, 5, feedbackValues, GL_INTERLEAVED_ATTRIBS);
    glLinkProgram(shaderParticles);
    GlShaderMgr::check_link_status(shaderParticles);

    // Free up shader compilation resources that are no longer needed.
    //GlShaderMgr::FinalizeCompileAndLink();

    RunSim_loc = glGetUniformLocation(shaderParticles, "RunSim");
    mousePos_loc = glGetUniformLocation(shaderParticles, "mousePos");
    ParticlePointSize_loc = glGetUniformLocation(shaderParticles, "ParticlePointSize");

    glUseProgram(shaderParticles);
    glUniform1i(RunSim_loc, 0);
    glUniform2f(mousePos_loc, 0.0f, 0.0f);

    // Set point size
    int i = 9;
    for (int NN = N; i > 1 && NN > 0; i--) {
        NN = NN >> 1;
    }
    glUniform1i(ParticlePointSize_loc, i);
         
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

// *******************************************************
// Process all key press events.
// This routine is called each time a key is pressed or released.
// *******************************************************
void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
	if (action == GLFW_RELEASE) {
		return;			// Ignore key up (key release) events
	}
    if (key == GLFW_KEY_ESCAPE || key == GLFW_KEY_X) {
        glfwSetWindowShouldClose(window, true);
        return;
    }
    if (key == GLFW_KEY_SPACE) {
        if (singleStep) {			// If ending single step mode
            singleStep = false;
            spinMode = true;		// Restart animation
        }
        else {
            spinMode = !spinMode;	// Toggle animation on and off.
        }
    }
    else if (key == GLFW_KEY_S) {
        singleStep = true;
        spinMode = true;
    }
    else if (key == GLFW_KEY_R) {
        singleStep = false;
        spinMode = false;
        firstTimeRendering = true;
        ResetParticles();
    }
    else {
        return;
    }
    glUniform1i(RunSim_loc, (spinMode || singleStep) ? 1 : 0);
}


// This function is called with the graphics window is first created,
//    and again whenever it is resized.
void window_size_callback(GLFWwindow* window, int width, int height) {
	glViewport(0, 0, width, height);		// Draw into entire window
    windowWidth = width;
    windowHeight = height;
}

// This function is called whenever the mouse position moves
void cursor_pos_callback(GLFWwindow* window, double x, double y) {

    // Get mouse position, scaled to range [-1,1]x[-1,1]
    float dotX = (2.0f*(float)x / (float)(windowWidth - 1)) - 1.0f;
    float dotY = 1.0f - (2.0f*(float)y / (float)(windowHeight - 1));

    glUniform2f(mousePos_loc, dotX, dotY);
}

void my_setup_OpenGL() {
    glEnable(GL_DEPTH_TEST);
    //glDepthFunc(GL_LESS);
    glEnable(GL_PROGRAM_POINT_SIZE);

	// The following commands may induce OpenGL to create round points and 
	//	antialias points.  (This is implementation dependent unfortunately.)
	//glEnable(GL_POINT_SMOOTH);
	//glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);	// Make round points, not square points
    // Comment out next lines to not have blending around points. 
    //    This is needed when points are too tightly packed, since
    //    otherwise the border antialiasing hides the points!
	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

void setup_callbacks(GLFWwindow* window) {
	// Set callback function for resizing the window
	glfwSetFramebufferSizeCallback(window, window_size_callback);

	// Set callback for key up/down/repeat events
	glfwSetKeyCallback(window, key_callback);

	// Set callbacks for mouse movement (cursor position) and mouse botton up/down events.
	glfwSetCursorPosCallback(window, cursor_pos_callback);
	// glfwSetMouseButtonCallback(window, mouse_button_callback);
}

// **********************
// Here is the main program
// **********************

int main() {
	glfwInit();
#if defined(__APPLE__) || defined(__linux__)
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
#endif
	const int initWidth = 800;
	const int initHeight = 600;
	GLFWwindow* window = glfwCreateWindow(initWidth, initHeight, "ParticlesTransformFeedback", NULL, NULL);
	if (window == NULL) {
		fprintf(stdout, "Failed to create GLFW window!\n");
		glfwTerminate();
		return -1;
	}
	glfwMakeContextCurrent(window);

	glewInit();

	// Print info of GPU and supported OpenGL version
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("OpenGL version supported %s\n", glGetString(GL_VERSION));
#ifdef GL_SHADING_LANGUAGE_VERSION
	printf("Supported GLSL version is %s.\n", (char *)glGetString(GL_SHADING_LANGUAGE_VERSION));
#endif
    printf("Using GLEW version %s.\n", glewGetString(GLEW_VERSION));

	printf("------------------------------\n");
    printf("%d many particles.\n", N*N);
    printf("Press SPACE to toggle(off and on) running the animation.\n");
    printf("Press 's' or 'S' to single-step animation.\n");
    printf("Press 'r' or 'R' to reset the particles.\n");
    printf("Press ESCAPE to exit.\n");
	
    setup_callbacks(window);
    window_size_callback(window, initWidth, initHeight);

	// Initialize OpenGL, the scene and the shaders
    my_setup_OpenGL();
	my_setup_SceneData();
 
    // Loop while program is not terminated.
	while (!glfwWindowShouldClose(window)) {
	
		myRenderScene();				// Render into the current buffer
		glfwSwapBuffers(window);		// Displays what was just rendered (using double buffering).

		// Poll events (key presses, mouse events)
        glfwWaitEventsTimeout(1.0/60.0);	// Use this to animate at 60 frames/sec (timing is NOT reliable)
	}

	glfwTerminate();
	return 0;
}

// If an error is found, it could have been caused by any command since the
//   previous call to check_for_opengl_errors()
// To find what generated the error, you can try adding more calls to
//   check_for_opengl_errors().
char errNames[8][36] = {
    "Unknown OpenGL error",
    "GL_INVALID_ENUM", "GL_INVALID_VALUE", "GL_INVALID_OPERATION",
    "GL_INVALID_FRAMEBUFFER_OPERATION", "GL_OUT_OF_MEMORY",
    "GL_STACK_UNDERFLOW", "GL_STACK_OVERFLOW" };
bool check_for_opengl_errors() {
    int numErrors = 0;
    GLenum err;
    while ((err = glGetError()) != GL_NO_ERROR) {
        numErrors++;
        int errNum = 0;
        switch (err) {
        case GL_INVALID_ENUM:
            errNum = 1;
            break;
        case GL_INVALID_VALUE:
            errNum = 2;
            break;
        case GL_INVALID_OPERATION:
            errNum = 3;
            break;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            errNum = 4;
            break;
        case GL_OUT_OF_MEMORY:
            errNum = 5;
            break;
        case GL_STACK_UNDERFLOW:
            errNum = 6;
            break;
        case GL_STACK_OVERFLOW:
            errNum = 7;
            break;
        }
        printf("OpenGL ERROR: %s.\n", errNames[errNum]);
    }
    return (numErrors != 0);
}