// *************
// vs1_PFT - First vertex shader for ParticlesFeedbackTransform
// *************
#beginglsl vertexshader vs1_PTF
#version 330 core
layout (location = 0) in vec2 vertPos;	   // Position in attribute location 0
layout (location = 1) in vec2 vertVel;     // Velocity in attribute location 1
layout (location = 2) in int vertexID;     // Vertex ID (particle) or -1 (mouse) in attribute location 2

// Uniform inputs
uniform int RunSim;							// ==1 (to run simulation), ==0 otherwise
uniform vec2 mousePos;						// Position of the mouse
uniform int ParticlePointSize;				// Size of particles

// Feedback transform outputs
out float posX, posY;					// Updated particle position
out float velX, velY;					// Updated particle velocity
out int vertID;						    // Vertex ID - passed thru

// Render output
out vec3 theColor;					       // Output a color to the fragment shader

float gravity = 0.2;
float deltaT = 1.0/60.0;
void CalcMotion();

void main()
{
   float posZ;
   if ( vertexID == -1 ) {
       theColor = vec3(1,1,1);  // Mouse point is white
	   posZ = -1.0;				// Mouse point positioned in front
	   gl_PointSize = 10;		// Mouse dot is larger
       posX = mousePos.x;
	   posY = mousePos.y;
   }
   else {
       theColor = vec3(1,0,0);
	   posZ = 0.0;
	   gl_PointSize = ParticlePointSize;
       if ( RunSim!=0 ) {
           CalcMotion();
       }
	   else { 
		   posX = vertPos.x;
		   posY = vertPos.y;
		   velX = vertVel.x;
		   velY = vertVel.y;
	   }
   }
   vertID = vertexID;
   gl_Position = vec4(posX, posY, posZ, 1.0);

   // The condition vertexID+1 == gl_VertexID will always hold! 
   // This is a somewhat unfortunate, but perhaps useful, feature of transform feedback.
   // Namely the output from transform feedback reflects the input order.
   // In this shader, vertexID is never used for anything.  In a situation
   //     where vertices are removed, vertexID would equal the original
   //     vertex index.
}

// Use the Euler-Cromer update method (symplectic) for
//    inverse square law gravitational force.
void CalcMotion() {
	vec2 displace = mousePos-vertPos;
	float distSq = max(displace.x*displace.x + displace.y*displace.y, 0.01);
	float dist = sqrt(distSq);
	vec2 accel = gravity*displace/(distSq*dist);
	vec2 newVel = vertVel + deltaT*accel;
	vec2 newPos = vertPos + deltaT*newVel;
	posX = newPos.x;
	posY = newPos.y;
	velX = newVel.x;
	velY = newVel.y;
}
#endglsl

// Fragment shader: fragmentShader_ColorOnly
//    Just passes thru the color value unchanged.
#beginglsl FragmentShader fragmentShader_ColorOnly
#version 330 core
in vec3 theColor;		// Color value came from the vertex shader (smoothed) 
out vec4 FragColor;	    // Color that will be used for the fragment
void main()
{
   FragColor = vec4(theColor, 1.0f);   // Add alpha value of 1.0.
}
#endglsl

